    processor 6502
    include vcs.h

;===============================================================================
; C O N S T A N T S
;===============================================================================
NUMROWS     =  5
NUMCOLUMS   = 11
NUMMOVES    =  8
INVADERCOL0 = $48
INVADERCOL1 = $ca


BIT_B       = $24
STA_ZP      = $85

RAMKernel   = $80

;===============================================================================
; Z P - V A R I A B L E S
;===============================================================================
    SEG.U   variables
    ORG     RAMKernel+$40  ; 64 bytes reserved for RAM kernel

tmpVar      .byte
tmpVar2     .byte
tmpColumn   = tmpVar2
FrameCnt    .byte
row         .byte
column      .byte
walkDir     .byte
rowCount    .byte

yStart      .byte

LinePat0    .byte           ; 11 bit required here!
LinePat1    .byte
LinePat2    .byte
LinePat3    .byte
LinePat4    .byte

;===============================================================================
; M A C R O S
;===============================================================================
; some delay macors with illegal NOP opcodes:
  MAC NOP_2
    .byte   $80,$00
  ENDM

  MAC NOP_3
    .byte   $04,$00
  ENDM

  MAC NOP_4
    .byte   $14,$00
  ENDM

;===============================================================================
; R O M - C O D E
;===============================================================================
    SEG     Bank0
    ORG     $f000

;TODO: replace sta.w with sta ZP,x to save bytes!

KernelCode:
.loopRAMKernel:
KernelPatch1:
    sta     RESP0           ; 3 +66         1st invader
;    bit     RESP0
    lda     #%10            ; 2             the 3rd copy is displayed
;    lda     #%1
    sta     NUSIZ0          ; 3 +71
    sta     HMOVE           ; 3 +74         blank used to mask copy 2 of 1st invader
;    bit     HMOVE
    NOP_2                   ; 2
    NOP_3                   ; 3
InvaderPatch:
    lda     Invader2Pat0,y  ; 4
    sta     GRP0            ; 3
    sta     GRP1            ; 3
    lda     #%1             ; 2             show 2 copies close (RESPx is patched to NUSIZx)
KernelPatch2:
    stx     NUSIZ0          ; 3             show 3 copies close (x=%11)
;    stx     RESP0
    sta.w   RESP1           ; 4 @22..30     2nd invader
    sta.w   RESP0           ; 4
    sta.w   RESP1           ; 4
    sta.w   RESP0           ; 4
    sta.w   RESP1           ; 4
    sta.w   RESP0           ; 4 @42..50
    sta.w   RESP1           ; 4
    sta.w   RESP0           ; 4
    sta.w   RESP1           ; 4
    sta.w   RESP0           ; 4 @58..66     11th invader
    dey                     ; 2
    bpl     .loopRAMKernel  ; 2
    jmp     NextRow
KernelCodeEnd:


Start SUBROUTINE
    sei                         ;           Disable interrupts, if there are any.
    cld                         ;           Clear BCD math bit.
    ldx     #0
    txs
    pha                         ;           Set stack to beginning.
    txa
.clearLoop:
    pha
    dex
    bne     .clearLoop

    jsr     GameInit

.mainLoop:
    jsr     VerticalBlank
    jsr     GameCalc
    jsr     DrawScreen
    jsr     OverScan
    jmp     .mainLoop


GameInit SUBROUTINE
    lda     #NUMROWS
    sta     row

;***** copy kernel to RAM: *****
    ldx     #KernelCodeEnd -KernelCode
.loopCopy:
    lda     KernelCode ,x
    sta     RAMKernel,x
    dex
    bpl     .loopCopy

    lda     #%01011101
    sta     LinePat0
    lda     #%10010010
    sta     LinePat1
    lda     #%01100101
    sta     LinePat2
    lda     #%11011111
    sta     LinePat3
    lda     #%10000110
    sta     LinePat4

    lda     #INVADERCOL1
    sta     COLUP0
    lda     #INVADERCOL1
    sta     COLUP1

    lda     #1
    sta     CTRLPF
    lda     #$70
    sta     PF0
    rts


VerticalBlank SUBROUTINE
    lda     #2
    sta     WSYNC
    sta     VSYNC
    sta     WSYNC
    sta     WSYNC
    lsr
    ldx     #44
    sta     WSYNC
    sta     VSYNC
    stx     TIM64T
    rts


GameCalc SUBROUTINE
    lda     SWCHB
    lsr
    bcs     .noReset
    jmp     Start
.noReset:

    lda     #$00
    bit     INPT4           ;       show how some tricky code works
    bmi     .noButton
    lda     #$02
.noButton:
    sta     COLUPF

    dec     FrameCnt
    bpl     .skipMove
    lda     #55+1           ;       number of invaders?
    lsr
    lsr
    lsr
    sta     FrameCnt

; let the formation walk:
    ldy     row
    iny
    cpy     #NUMROWS
    bcc     .skipColumnMove
    ldy     #0
    ldx     column
    bit     walkDir
    bmi     .leftDir
    inx
    cpx     #NUMMOVES
    bcs     .invertDir
    bcc     .skipInvert

.leftDir:
    dex
    bne     .skipInvert

.invertDir:
;    ldy     #NUMROWS        ;       only move down
    lda     walkDir
    eor     #$80
    sta     walkDir
    inc     yStart
    lda     yStart
    cmp     #8
    bcc     .skipInvert
    lda     #0
    sta     yStart
.skipInvert:
    stx     column
.skipColumnMove:
    sty     row
.skipMove:
    rts

    align 256

DrawScreen SUBROUTINE
;    tsx
;    stx     tmpVar

    ldx     #228
.waitTim:
    lda     INTIM
    bne     .waitTim
    sta     WSYNC
    sta     VBLANK
    stx     TIM64T
;---------------------------------------------------------------
    lda     #%1
    sta     NUSIZ1

;***** wait some rows: *****
    ldy     yStart
    ldx     YStartTab,y
.loopWSYNC:
    sta     WSYNC
    dex
    bpl     .loopWSYNC

    lda     column
    sta     tmpColumn
    ldx     #NUMROWS-1
    jmp     .enterKernel

.exitLoopJmp:
    jmp     .exitLoop

NextRow:
    ldy     #0
    sty     GRP0
    sty     GRP1

    ldx     rowCount
    dex
    bmi     .exitLoopJmp
.enterKernel:
    stx     rowCount
;***** update tmpColumn: *****
    cpx     row                         ; change column +/- 1
    bne     .skipColumn
    lda     walkDir
    bmi     .negDir
    inc     tmpColumn
    bne     .skipColumn
.negDir:
    dec     tmpColumn
.skipColumn:

;***** set HMP0: *****
    ldy     #0
    lda     tmpColumn
    bne     .notFirstCol
    ldy     #$80                        ; early HMOVE (at 74)
.notFirstCol:
    sty     HMP0

;***** animate enemies: *****
    lsr
    ldy     InvaderTabLo,x
    bcc     .noInx
    ldy     InvaderTabLo+NUMROWS,x
.noInx:
    sty.b   RAMKernel+InvaderPatch+1-KernelCode,x

    sta     WSYNC

;***** patch kernel (part 1/2): *****
    lda     LinePat0,x
    sta     tmpVar
    ldx     #3*8-3
.loopEnable:
    lsr     tmpVar
    ldy     #RESP0
    bcs     .enabled
    ldy     #NUSIZ0
.enabled:
    txa
    lsr
    bcs     .player0
    iny
.player0:
    sty     RAMKernel+KernelPatch2+3-KernelCode,x
    dex
    dex
    dex
    bpl     .loopEnable

    sta     WSYNC
;---------------------------------------
;***** patch kernel (part 2/2): *****
    lda     tmpColumn       ; 3
    cmp     #4              ; 2
    bcs     .rightColumn    ; 2
    lda     #STA_ZP         ; 2
    ldx     #%10            ; 2
    ldy     #NUSIZ0         ; 2
    bne     .patchCode      ; 3
.rightColumn:
    lda     #BIT_B          ; 2
    ldx     #%1             ; 2
    ldy     #RESP0          ; 2
    nop                     ; 2
.patchCode:
    sta     KernelPatch1   + [RAMKernel-KernelCode]; 3
    sta     KernelPatch1+6 + [RAMKernel-KernelCode]; 3
    stx     KernelPatch1+3 + [RAMKernel-KernelCode]; 3
    sty     KernelPatch2+1 + [RAMKernel-KernelCode]; 3 = 28

    NOP_4
    NOP_2

;***** delay kernel start: *****
    lda     tmpColumn       ; 3
    lsr                     ; 2
    bcs     .delayC1        ; 2
.delayC1:
    lsr                     ; 2
    bcc     .delayC2        ; 2
    NOP_3                   ; 3
.delayC2:
    lsr                     ; 2
    bcc     .delayC4        ; 2
    NOP_3                   ; 3
    NOP_2                   ; 2
.delayC4:
    lsr                     ; 2
    bcc     .delayC8        ; 2
    NOP_4                   ; 4
    NOP_3                   ; 3
    NOP_2                   ; 2
.delayC8:                   ;     = 22..30

    ldy     #8              ; 2
    ldx     #%11            ; 2   =  4

    jmp     RAMKernel       ; 3
;---------------------------------------------------------------

.exitLoop:

;---------------------------------------------------------------
    ldx     #2
.waitScreen:
    lda     INTIM
    bne     .waitScreen
    sta     WSYNC
    stx     VBLANK

;    ldx     tmpVar
;    txs
    rts

OverScan SUBROUTINE
    lda     #36
    sta     TIM64T

.waitTim:
    lda     INTIM
    bne     .waitTim
    rts


;===============================================================================
; R O M - T A B L E S
;===============================================================================
    org     $f700

InvaderTabLo:
    .byte   <Invader0Pat0, <Invader0Pat0, <Invader1Pat0, <Invader1Pat0, <Invader2Pat0
    .byte   <Invader0Pat1, <Invader0Pat1, <Invader1Pat1, <Invader1Pat1, <Invader2Pat1

SpriteTab0:
Invader0Pat0:
    .byte   %10000001
    .byte   %01011010
    .byte   %01100110
    .byte   %11111111
    .byte   %10011001
    .byte   %11111111
    .byte   %01111110
    .byte   %00011000
    .byte   0
SpriteTab:
Invader0Pat1:
    .byte   %01000010
    .byte   %10011001
    .byte   %01100110
    .byte   %11111111
    .byte   %10011001
    .byte   %11111111
    .byte   %01111110
    .byte   %00011000
    .byte   0

Invader1Pat0:
    .byte   %00011000
    .byte   %10100101
    .byte   %10111101
    .byte   %11111111
    .byte   %01011010
    .byte   %00111100
    .byte   %00100100
    .byte   %01000010
    .byte   0
Invader1Pat1:
    .byte   %01000010
    .byte   %00100100
    .byte   %00111100
    .byte   %01111110
    .byte   %11011011
    .byte   %10111101
    .byte   %10100101
    .byte   %01000010
    .byte   0

Invader2Pat0:
    .byte   %00100100
    .byte   %01000010
    .byte   %00100100
    .byte   %01111110
    .byte   %01011010
    .byte   %01111110
    .byte   %00111100
    .byte   %00011000
    .byte   0
Invader2Pat1:
    .byte   %01000010
    .byte   %00100100
    .byte   %00011000
    .byte   %01111110
    .byte   %01011010
    .byte   %01111110
    .byte   %00111100
    .byte   %00011000
    .byte   0

YStartTab:
    .byte   0,15,30,45,60,75,90,105

    .byte   " Tom's-Invader-Demo v0.1 - Copyright (C) 2001 Thomas Jentzsch "

    org     $f7fc
    .word   Start
    .word   Start
